#/usr/bin/env python
# -*- coding: utf-8 -*-

# Aufrufen mit: execfile('UTF-8-escape.py');

# (c) Jonas Bechtel 2017-2018

# Thanks to https://stackoverflow.com/questions/11624190/python-convert-string-to-byte-array
# My method: b = bytearray(string)
#            s = str(b)
# Thanks to https://de.wikipedia.org/wiki/UTF-8


# Testcase: (Nach Belieben B fett, I kursiv und U unterstrichen machen.)
# 'äB -> UTF8-2-195164
# '€' -> UTF8-3-226130172
# B–' -> UTF8-3-226128147
# I𝄞' -> UTF8-4-240157132158
# '€U -> UTF8-3-226130172
# 'ä' -> UTF8-2-195164
# '𝄞' -> UTF8-4-240157132158

"""
# Testcode:
path_name='/home/knoppix/compile/WiScri/Minitest'
f = open(path_name + '/UTF-8-escape.py'); exec(f.read()); f.close();

utf_check_16bit_constraint('Supertoll'.encode('utf-8')) # -> true
utf_check_16bit_constraint('asdgewähltf𝄞fee'.encode('utf-8')) # -> false
o_utf_check_16bit_constraint() # -> depends on selected object.


scribus.selectText(25,4, "TO")
v=scribus.getAllText()
v
v.encode('utf-8')
scribus.deselectAll()
scribus.selectObject("TO")
scribus.selectText(50,5)
v2=scribus.getAllText()
v2
v2.encode('utf-8')

o_utf_escape("TO", 4)
o_back_to_utf("TO", 2) # ,3)  #  ,4)
""";

# pure text (bytearray)
def utf_check_16bit_constraint(b_in, lvl=4):
	i_quit = len(b_in);
	i = 0;
	OK = 1;
	while (i < i_quit):
		b = b_in[i];
		if (194 <= b and b <= 223):
			if lvl <= 2:
				OK=0;
				break;
			i += 1;
		elif (224 <= b and b <= 239):
			if lvl <= 3:
				OK=0;
				break;
			i += 2;
		elif (240 <= b and b <= 247):
			# four-byte UTF-8 -> causes wrong counts with scribus.
			if lvl <= 4: # Must be true. lvl > 4 makes no sense.
				OK=0;
				break;
			i += 3;
			
		elif (b < 128):
			# ASCII
			nop = 0; # = continue
			#continue;
		else:
			print ('UTF-8-escape: utf_escape(): WARNING: ignoring unknown/unsupported byte {} @in:{}'.format(b, i));
		
		i += 1;
	
	if (OK == 0):
		print ('UTF-8-escape: o_utf_check_16bit_constraint(): WARNING: Checked text is NOT clean! (byte {} @ {})'.format(int(b), i));
	
	return OK;


# Scribus object
def o_utf_check_16bit_constraint(name=[], lvl=4):
	if name == []:
		text_in = scribus.getAllText();
	else:
		text_in = scribus.getAllText(name);
	b_in = text_in.encode('utf-8');
	return utf_check_16bit_constraint(b_in, lvl);



# pure text
def utf_escape(text_in):
	text_out = '';
	b_in = bytearray(text_in);
	
	#p_out = 0;
	p_last = 0;
	i_quit = len(text_in)
	i = 0;
	while (i < i_quit):
		b = b_in[i];
		if (194 <= b and b <= 223):
			if (p_last != i):
				text_out += str(b_in[p_last:i]);
	#			p_out += i - p_last;
				p_last = i + 2;
			text_out += '``UTF8-2-' + str(b) + str(b_in[i+1]);
			i += 1;
			p_last = i + 1;
		elif (224 <= b and b <= 239):
			if (p_last != i):
				text_out += str(b_in[p_last:i]);
	#			p_out += i - p_last;
				p_last = i + 2;
			text_out += '``UTF8-3-' + str(b) + str(b_in[i+1]) + str(b_in[i+2]);
			i += 2;
			p_last = i + 1;
			
		elif (b < 128):
			# ASCII
			trash = 0; # = continue
			#continue;
		else:
			print ('UTF-8-escape: utf_escape(): WARNING: ignoring unknown/unsupported byte {} @in:{}/out:{}'.format(b, i, len(text_out)));
		
		i += 1;
	
	text_out += str(b_in[p_last:]);
	
	return text_out;

# Scribus object
def o_utf_escape(obj_name, lvl=4):
	scribus.selectText(1, 0, obj_name);
	scribus.insertText('#', -1, obj_name); # TODO: Beginning of the line, not beginning of the object is impotant!
	text_in = scribus.getAllText(obj_name);
	b_in = text_in.encode('utf-8');
	
	p_bias = 0;
	p_last = 0;
	last_linebreak = -1;
	i_quit = len(text_in)
	i = 0;
	while (i < len(b_in)): # (len is changing)
#		print ('{}: {}'.format(i, b_in[i]))
		b = b_in[i];
			
		if 1:
			linebreak = b_in.rfind(b'\r', 0, i);
			# Inserting and removing a newline will settle fonts
			# (circumventing Scribus Bug)
			# (only needed one time after copying, so we use it only in
			#		utf_escape and not in back_to_utf.)
			if linebreak == -1:
				scribus.insertText('X\rY', 0, obj_name);
				scribus.selectText(0,3,obj_name)
				scribus.deleteText(obj_name)
			elif linebreak != last_linebreak:
				# Die Indexierung funktioniert deswegen, weil nur 1x mal
				# auf den Zeilenumbruch zugegriffen wird. Bei den folgenden
				# Malen sparen wir das, was auch gut ist, denn ansonsten
				# wären die Indexe verschoben. (Würde auch funktionieren,
				# nür würde dann nicht immer exakt der Zeilenanfang
				# getroffen werden.)
				scribus.insertText('X\rY', linebreak+1 + p_bias, obj_name);
				scribus.selectText(linebreak+1 + p_bias-2,7,obj_name)
				#print str(i) + '  ' + str(linebreak) + '  ' + str(p_bias) + ' -> ' + scribus.getAllText(obj_name)
				scribus.selectText(linebreak+1 + p_bias,3,obj_name)
				scribus.deleteText(obj_name)
				last_linebreak = linebreak;
			
		if (b < 128):
			# ASCII control character
			1;
		elif (128 <= b and b <= 193):
			# ASCII normal character
			1;
		elif (194 <= b and b <= 223):
			if (lvl <= 2): # lvl < 2 makes no sense.
#				print ([119944, i, p_bias, 1])
				
				# You cannot do it right: when inserting left of the complex
				#   char, you get the underline, but not the rest of the style.
				#   When inserting right of the complex char, you've got the
				#   style, but no underlines. (same to UTF8-3 and UTF8-4)
				scribus.insertText('``UTF8-2-' + str(b) + str(b_in[i+1]), i + p_bias + 1, obj_name);
				addlen = 15;
				
				scribus.selectText(i + p_bias, 1, obj_name);
				tmp_trash = scribus.getAllText(obj_name);
				font = scribus.getFont(obj_name);
#				print ('{}+{} ({}) -> _{}_'.format(i + p_bias, 1, font, tmp_trash));
				scribus.deleteText(obj_name);
				
				scribus.selectText(i + p_bias, addlen, obj_name)
				scribus.setFont(font, obj_name);
				
				p_bias += addlen; # Added a long string.
				p_bias -= 1; # Taken one codepoint
				p_last = i + 2;
				#break;
			
			i += 1;
			p_bias -= 1; # We have used 1 additional byte from b_in which are 'transparent' for scribus.
		elif (224 <= b and b <= 239):
			if (lvl <= 3):
#				print ([222244, i, p_bias, 1])
				
				scribus.insertText('``UTF8-3-' + str(b) + str(b_in[i+1]) + str(b_in[i+2]), i + p_bias + 1, obj_name);
				addlen = 18;
				
				scribus.selectText(i + p_bias, 1, obj_name);
				font = scribus.getFont(obj_name);
#				print ('{}+{} -> _{}_'.format(i + p_bias, 1, scribus.getAllText(obj_name)));
				scribus.deleteText(obj_name);
				
				scribus.selectText(i + p_bias, addlen, obj_name)
				scribus.setFont(font, obj_name);
				
				p_bias += addlen; # Added a long string.
				p_bias -= 1; # Taken one codepoint
			
			i += 2;
			p_bias -= 2; # We have used 2 additional bytes from b_in which are 'transparent' for scribus.
		elif (240 <= b and b <= 247):
			if (lvl <= 4): # Must be true. lvl > 4 makes no sense.
#				print ([224400, i, p_bias, 1])
				scribus.insertText('``UTF8-4-' + str(b) + str(b_in[i+1]) + str(b_in[i+2]) + str(b_in[i+3]), i + p_bias + 2, obj_name);
				addlen = 21;
				
				scribus.selectText(i + p_bias, 2, obj_name);
				font = scribus.getFont(obj_name);
#				print ('{}+{} -> _{}_'.format(i + p_bias, 2, scribus.getAllText(obj_name)));
				scribus.deleteText(obj_name);
				
				scribus.selectText(i + p_bias, addlen, obj_name)
				scribus.setFont(font, obj_name);
				
				p_bias += addlen; # Added a long string.
				p_bias -= 2; # Taken one codepoint = 2 QChars = 2 UTF16 chars
			
			i += 3;
			p_bias -= 2; # We have used 2 additional bytes from b_in which are 'transparent' for scribus.
			
		else:
			print ('UTF-8-escape: utf_escape(): WARNING: ignoring unknown/unsupported byte {} @in:{}/out:{}'.format(b, i, len(text_in)));
		
		i += 1;
	
	#print [i, 'p_bias:', p_bias, -1]
	scribus.selectText(i + p_bias - 1, 1, obj_name);
#	print ('{}+{} -> {}'.format(i + p_bias - 1, 1, scribus.getAllText(obj_name)));
	scribus.deleteText(obj_name);
	

# pure text
def back_to_utf(esc_in):
	text_out = '';
	
	p_last = 0;
	while True:
		pu = esc_in.find('``UTF8-', p_last);
		#print pu;
		if (pu == -1):
			break;
		if (pu != p_last):
			text_out += esc_in[p_last:pu];
		pm = pu + 7;
		if esc_in[pm:pm+2] == '2-':
			i1 = int(esc_in[pm+2:pm+5]);
			i2 = int(esc_in[pm+5:pm+8]);
		#	print [1, i1, i2]
			nstr = str(bytearray([i1, i2]));
			p_last = pm + 8;
			text_out += nstr;
		elif esc_in[pm:pm+2] == '3-':
			i1 = int(esc_in[pm+2:pm+5]);
			i2 = int(esc_in[pm+5:pm+8]);
			i3 = int(esc_in[pm+8:pm+11]);
		#	print [1, i1, i2, i3]
			nstr = str(bytearray([i1, i2, i3]));
			p_last = pm + 11;
			text_out += nstr;
		else:
			p_last = pm;
			continue;
	if (p_last != len(esc_in)):
		text_out += esc_in[p_last:];
	return text_out;


# on scribus object
def o_back_to_utf(obj_name, lvl=2):
	scribus.selectText(0, 0, obj_name);
	scribus.insertText('#', -1, obj_name);
	esc_in = scribus.getAllText(obj_name);
	origlen = len(esc_in);
	# Directly operating on utf-8 string.
	#esc_in = esc_in.encode('utf-8');
	
	# Wir rollen das Feld von hinten auf.
	# We process from end to beginning -> very easy.
	
	p_bias = 0;
	p_first = origlen;
	last_linebreak = -1;
	while p_first >= 0:
		pu = esc_in.rfind('``UTF8-', 0, p_first);
		#print pu;
		if pu == -1:
			break;
		
			# (only needed one time after copying, so we use it only in
			#		utf_escape and not in back_to_utf.)
		#linebreak = esc_in.rfind('\r', 0, pu);
		#if linebreak == -1:
		#	scribus.insertText('X\rY', 0, obj_name);
		#	scribus.selectText(0,3,obj_name)
		#	scribus.deleteText(obj_name)
		#elif linebreak != last_linebreak:
		#	scribus.insertText('x\rY', linebreak, obj_name);
		#	scribus.selectText(linebreak,3,obj_name)
		#	scribus.deleteText(obj_name)
		#	last_linebreak = linebreak;
		
		pm = pu + 7;
		p_first = pu-3;
		if lvl <= 2 and esc_in[pm:pm+2] == '2-':
			i1 = int(esc_in[pm+2:pm+5]);
			i2 = int(esc_in[pm+5:pm+8]);
			#print ([1, i1, i2])
			nstr = bytearray([i1, i2]).decode('utf-8');
			#print (['pm:', pm, 2, 6])
			scribus.insertText(nstr, pm + 2+6, obj_name);
			scribus.selectText(pu, 7+ 2+6, obj_name);
			font = scribus.getFont(obj_name)
			#print ('{}+{} -> {}'.format(i + p_bias, 1, scribus.getAllText(obj_name)));
			scribus.deleteText(obj_name);
			scribus.selectText(pu, 1, obj_name)
			scribus.setFont(font, obj_name);
			p_bias += 1 - (7+ 2+6);
		elif lvl <= 3 and esc_in[pm:pm+2] == '3-':
			i1 = int(esc_in[pm+2:pm+5]);
			i2 = int(esc_in[pm+5:pm+8]);
			i3 = int(esc_in[pm+8:pm+11]);
			#print ([1, i1, i2, i3])
			nstr = bytearray([i1, i2, i3]).decode('utf-8');
			#print (['pm:', pm, 2, 9])
			scribus.insertText(nstr, pm + 2+9, obj_name);
			scribus.selectText(pu, 7+ 2+9, obj_name);
			font = scribus.getFont(obj_name)
			#print '{}+{} -> {}'.format(i + p_bias, 1, scribus.getAllText(obj_name));
			scribus.deleteText(obj_name);
			scribus.selectText(pu, 1, obj_name)
			scribus.setFont(font, obj_name);
			p_bias += 1 - (7+ 2+9);
		elif lvl <= 4 and esc_in[pm:pm+2] == '4-':
			i1 = int(esc_in[pm+2:pm+5]);
			i2 = int(esc_in[pm+5:pm+8]);
			i3 = int(esc_in[pm+8:pm+11]);
			i4 = int(esc_in[pm+11:pm+14]);
			#print ([1, i1, i2, i3, i4])
			nstr = bytearray([i1, i2, i3, i4]).decode('utf-8');
			#print (['pm:', pm, 2, 12])
			scribus.insertText(nstr, pm + 2+12, obj_name);
			scribus.selectText(pu, 7+ 2+12, obj_name);
			font = scribus.getFont(obj_name)
			#print '{}+{} -> {}'.format(i + p_bias, 1, scribus.getAllText(obj_name));
			scribus.deleteText(obj_name);
			scribus.selectText(pu, 2, obj_name)
			scribus.setFont(font, obj_name);
			p_bias += 2 - (7+ 2+12);
	
	#print ([origlen, 'p_bias:', p_bias, -1]);
	scribus.selectText(origlen + p_bias - 1, 1, obj_name);
	#print ('{}+{} -> {}'.format(origlen + p_bias - 1, 1, scribus.getAllText(obj_name)));
	scribus.deleteText(obj_name);
	
